#!/usr/bin/perl -w -T
########################################################
#
# This script is intended to provide a means for
# detecting changes made to files, via a regular
# comparison of MD5 hashes to an established
# "baseline".
# In this respect, it is designed as a portable clone
# of aide software.
#
# This script requires perl ,and some others perl modules
# which come in standard installation
#
###############################################################################
#    Copyright (C) 2002 by Eric Gerbier
#    Bug reports to: gerbier@users.sourceforge.net
#    $Id: afick.pl,v 1.52 2004/01/28 13:09:32 gerbier Exp $
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################

use strict;

use Digest::MD5;    # for md5 checksum

# can also use Digest::SHA1 : see below

use POSIX qw(:sys_stat_h);    # for file type macros

use SDBM_File;                # database
use FileHandle;               # for database open mode
use Getopt::Long;             # arg analysis
use Cwd 'abs_path';

use File::Glob ':glob';       # for jokers
use File::Basename;           # for delete

# !!!!
# Win32::FileSecurity may be used too
# or even better : Win32::Perms

################# global variables #############################

use vars qw ($Dbm);

$Dbm = 'SDBM_File';           # database type

use vars
  qw (%Finded %Hashfile %Newfiles %Dangling %Toscan %Onlydir %Onlythis %Oldval %Newval @ExSufx);
%Finded   = ();               # list of files scanned
%Hashfile = ();               # dbm hash-table
%Newfiles = ();               # list of new files
%Dangling = ();               # dangling links
%Toscan   = ();               # list to scan
%Onlydir  = ();               # only this and content, without sub-dir
%Onlythis = ();               # only this inode
%Oldval   = ();               # store old values
%Newval   = ();               # store new changes values
@ExSufx   = ();               # store excluded suffixes

use vars qw ( $Nbmod  $Update $File);
$Nbmod  = 0;                  # number of modifications
$Update = 0;                  # update/compare mode ?
$File   = 0;                  # file mode ?

# directives status
# not initialized to be able to detect negative command lines
use vars
  qw ($Verbose $Warn_dead_symlinks $Report_full_newdel $Warn_missing_file $Report_url $Progress $Timing $Running);

#$Verbose            ;           # verbose mode ?
#$Warn_dead_symlinks = '';           # warn_dead_symlink
#$Report_full_newdel     = '';	   # report new directories contents
#$Warn_missing_file   = '';
$Report_url = "STDOUT";       # report_url

my $version = "1.8-0";

# alias table
# contains pre-defined values
my %alias = (
	a    => "a",              # atime
	b    => "b",              # blocks
	c    => "c",              # ctime
	d    => "d",              # device
	g    => "g",              # gid
	i    => "i",              # inode
	m    => "m",              # mtime
	md5  => "5",              # md5 hash
	h    => "5",              # md5 hash by default (compatibility)
	sha1 => "1",              # sha1
	n    => "n",              # number of links
	p    => "p",              # permissions
	u    => "u",              # uid
	s    => "s",              # size
	all  => "pdinugsbmc5",    # all except a
	R    => "pdinugsmc5",     # aide compatibility
	L    => "pdinug",         # aide compatibility
	E    => "",               # aide compatibility
);
################ end of global var #############################

my @listerep;                 # list of rep to scan
my $database;                 # database name
my $acl_exist;
my $sha1_exist;

# command line arguments/flags
my $configfile;               # config file name
my $help;
my $init;
my $compare;
my $list;
my $print;
my $Version;
my $Sufx_list;
my $update;

my $database_ext = ".pag";    # database extension
my $control_ext  = ".ctr";    # database control file

my $default_config_file;
my $default_alias = "all";

# index hash for internal info
my %id = (
	'md5'       => 0,
	'device'    => 1,
	'inode'     => 2,
	'file_mode' => 3,
	'links'     => 4,
	'uid'       => 5,
	'acl'       => 6,    # acl data are same than gid
	'gid'       => 6,
	'filesize'  => 7,
	'blocs'     => 8,
	'atime'     => 9,
	'mtime'     => 10,
	'ctime'     => 11
);

# array of field names
my @field = (
	'checksum  ',        # 0
	'device    ',        # 1
	'inode     ',        # 2
	'file mode ',        # 3
	'links     ',        # 4
	'uid       ',        # 5
	'gid       ',        # 6
	'filesize  ',        # 7
	'blocs     ',        # 8
	'atime     ',        # 9
	'mtime     ',        # 10
	'ctime     '         # 11
);

# database internal separator
my $sep     = '|';                # for join
my $sepmeta = quotemeta($sep);    # for split

my $sepacl  = ',';
my $comment = '#';
#######################################################
# a general methode to access database data
sub get_data($$) {

	my $name  = shift (@_);       # file name
	my $field = shift (@_);

	my $hash = $Hashfile{$name} || $Newval{$name} || "";
	my @tab = split ( /$sepmeta/, $hash );
	return $tab[ $id{$field} ];
}
#######################################################
# get checksum from database
sub get_md5 ($) {
	my $name = shift (@_);        # file name
	return get_data( $name, 'md5' );
}
#######################################################
# get inode date from database
sub get_ctime ($) {
	my $name = shift (@_);        # file name

	return get_data( $name, 'ctime' );
}
#######################################################
# get inode date from database
sub get_mtime ($) {
	my $name = shift (@_);        # file name

	return get_data( $name, 'mtime' );
}
#######################################################
# return true if a socket
# same format as posix macros
sub S_IFSOCK($) {
	my $p = shift (@_);

	# see sys/stat.h
	my $S_IFSOCK = 0140000;

	return ( ( $p & $S_IFSOCK ) == $S_IFSOCK );
}
#######################################################
# return true if a link
# same format as posix macros
sub S_IFLNK($) {
	my $p = shift (@_);

	# see sys/stat.h
	my $S_IFLNK = 0120000;

	return ( ( $p & $S_IFLNK ) == $S_IFLNK );
}
#######################################################

# get file_type from database
sub file_type ($) {
	my $name = shift (@_);    # file name

	# $Hashfile for changed file
	# Newval in compare mode for new files
	my $hash;
	if ( exists $Hashfile{$name} ) {
		$hash = $Hashfile{$name};
	}
	elsif ( exists $Newval{$name} ) {
		$hash = $Newval{$name};
	}
	else {
		warning("file_type problem on $name\n");
		return "unknown_type";
	}

	my @tab  = split ( /$sepmeta/, $hash );
	my $perm = $tab[ $id{'file_mode'} ];

	# now use POSIX macro if exists
	# else my own sub  S_IFLNK S_IFSOCK

	my $type;

	# order of test is important for the first test !
	if ( S_IFSOCK($perm) ) {
		$type = "socket";
	}
	elsif ( S_IFLNK($perm) ) {
		$type = "symbolic_link";
	}
	elsif ( S_ISREG($perm) ) {
		$type = "file";
	}
	elsif ( S_ISBLK($perm) ) {
		$type = "block_device";
	}
	elsif ( S_ISDIR($perm) ) {
		$type = "directory";
	}
	elsif ( S_ISCHR($perm) ) {
		$type = "character_device";
	}
	elsif ( S_ISFIFO($perm) ) {
		$type = "fifo";
	}
	else {
		$type = "unknown_type";
	}
	return $type;
}
#######################################################
# low-level sub to send "normal" messages
sub report($) {
	no strict 'refs';
	print $Report_url @_ if ( $Report_url ne 'null' );
}
#######################################################
# high-level sub for summary messages
sub report_summary($$) {
	my $type     = shift (@_);    # new, change, deleted, dangling ...
	my $filename = shift (@_);

	my $filetype = file_type($filename);

	report("$type $filetype : $filename\n");
}
#######################################################
# high-level sub for detailed messages
sub report_detailed($$;$) {
	my $field = shift (@_);
	my $val1  = shift (@_);
	my $val2  = shift (@_);

	report("\t$field\t\t : $val1");
	if ( defined $val2 ) {
		report("\t$val2\n");
	}
	else {
		report("\n");
	}
}
#######################################################
# print a commented line
sub info($) {
	report("$comment @_");
}
#######################################################
# for progress purpose
sub progress($) {
	print "progress @_\n";
}
#######################################################
# for debug purpose
sub debug($) {
	print "DEBUG: @_" if ($Verbose);
}
#######################################################
sub warning($) {
	my $text = shift (@_);
	warn "WARNING: $text";
}

#######################################################
# remove_trailing blanks
sub remove_trailing_spaces ($) {
	my $val = shift (@_);
	$$val =~ s/^\s+//;
	$$val =~ s/\s+$//;
}
#######################################################
# remove all spaces
sub remove_all_spaces ($) {
	my $val = shift (@_);
	$$val =~ s/\s+//g;
}
#######################################################
# test_double
# to detect bad configuration
sub test_double($) {
	my $elem = shift (@_);
	return exists $Toscan{$elem};
}
#######################################################
# used to exclude suffixes
sub test_exclude_suffix($) {
	my $fic = shift (@_);
	foreach my $sufx (@ExSufx) {
		if ( $fic =~ m/\.$sufx$/ ) {
			debug("find excluded suffix $sufx in $fic\n");
			return 1;
		}
	}
	return 0;
}
#######################################################
# expand_line
# expand config line : treat jokers, and split into <file attribute flagdir>
sub trait_line ($) {
	my $line = shift (@_);

	my $name;    # file name
	my $attribute = "";    # what to do on $name

	# for file with blank caracters
	if ( $line =~ m/^"(.*)"(.*)/ ) {
		$name      = $1;
		$attribute = $2;
	}
	else {
		( $name, $attribute ) = split ( " ", $line, 2 );
	}

	# for compatibilite with old syntaxe
	$attribute = $default_alias if ( !$attribute );

	# debug( "$name -> $attribute");

	my @tab;
	my $flagdir;

	# expand jokers if exists
	if ( $name =~ m/[\*\?]/ ) {
		@tab = glob($name);
		debug("expand $name in @tab\n");
	}
	elsif ( $name eq "/" ) {
		$flagdir = 1;
		push ( @tab, $name );
	}
	elsif ( $name =~ m?/$? ) {

		# remove trailing /
		$name =~ s?/$??;
		$flagdir = 1;
		push ( @tab, $name );
	}
	else {
		$flagdir = 0;
		push ( @tab, $name );
	}
	push ( @tab, $attribute, $flagdir );
	return @tab;
}
#######################################################
#resolv_alias
# expand and resolv action pattern, for alias definitions and files config
# be careful : not recursif
sub resolv_alias($$) {
	my $nline = shift (@_);    # line number in config file
	my $val   = shift (@_);    # value

	my $realval = "";

	remove_all_spaces( \$val );
	$val =~ m/^([^+-]*)/;      # first motif before any +/-
	my $first_alias = $1;
	debug("first alias : $first_alias\n");
	if ( !exists $alias{$first_alias} ) {

		# error : pattern not found
		warning(
"error in config file (line $nline : $val) : first alias \"$first_alias\" can not be resolved (ignored)\n"
		);
	}
	else {
		$realval = $alias{$first_alias};
	}

	# then other flags
	# first negative ones
	while ( $val =~ m/-([^+-]*)/g ) {
		my $remov_alias = $1;
		debug("soustractive alias $remov_alias\n");
		if ( !exists $alias{$remov_alias} ) {

			# remove bad aliases
			warning(
"error in config file (line $nline : $val) :  soustractive alias \"$remov_alias\" can not be resolved (ignored)\n"
			);
		}
		else {
			$realval =~ s/$remov_alias//;
		}
	}

	# then additives ones
	while ( $val =~ m/\+([^+-]*)/g ) {
		my $add_alias = $1;
		debug("additive alias $add_alias\n");
		if ( !exists $alias{$add_alias} ) {

			# remove bad aliases
			warning(
"error in config file (line $nline : $val) : additive alias \"$add_alias\" can not be resolved (ignored)\n"
			);
			next;
		}
		elsif ( $realval =~ m/$alias{$add_alias}/ ) {

			# do not allow duplicates
			warning(
"error in config file (line $nline : $val) : duplicate alias $add_alias in $realval\n"
			);
		}
		else {
			$realval .= $alias{$add_alias};
		}
	}

	if ( !$realval ) {
		$realval = $alias{$default_alias};
		warning(
			"set to default pattern \"$default_alias\" in config line line $.\n"
		);
	}

	# test for optionnal sha1
	if ( $realval =~ m/1/ ) {
		if ($sha1_exist) {

			# sha1 take place of md5
			if ( $realval =~ m/5/ ) {
				$realval =~ s/5//;
				warning(
"can not use sha1 and md5 (line $nline : $val) : keep sha1\n"
				);
			}
		}
		else {
			warning(
"sha1 is not installed, will be replaced by md5 checksum on config file (line $nline : $val)\n"
			);

			# a test to avoid duplicate
			if ( $realval !~ m/5/ ) {
				$realval =~ s/1/5/;
			}
		}
	}

	return $realval;

}
#######################################################
# decode
# used to read directives : yes/1/true or no/0/false

sub decode ($$$) {
	my $val = lc( shift (@_) );    # nocase
	my $line   = shift (@_);       # line
	my $nuline = shift (@_);       # number of line

	if ( ( $val eq "yes" ) or ( $val eq "true" ) or ( $val eq "1" ) ) {
		return 1;
	}
	elsif ( ( $val eq "no" ) or ( $val eq "false" ) or ( $val eq "0" ) ) {
		return 0;
	}
	else {
		warning("bad value $val in config line line $nuline : $line\n");

		# return as no
		return 0;
	}
}
#######################################################
# get_names
# used by auto_control to know which program names to control
sub get_names() {
	my ( $name, $path, undef ) = fileparse($0);

	$path = abs_path($path) . '/';    # convert in absolute path

	my @list;

	# because of links from afick.pl to afick
	my $afick = $path . "afick.pl";
	if ( -x $afick ) {
		push ( @list, $afick );
	}
	else {
		push ( @list, $path . $name );
	}

	my $aficktk = $path . "afick-tk.pl";

	# is afick-tk installed ?
	if ( -x $aficktk ) {
		push ( @list, $aficktk );
	}

	return @list;
}
#######################################################
# ignore_directive
# used by read_config for a common message
sub ignore_directive($$) {
	my $line    = shift (@_);
	my $numline = shift (@_);
	debug("ignore $line (line $numline) : overload by command line option\n");
}
#######################################################
# read_config :
# read the config file to build a list of directory to scan
sub read_config($) {
	my $config = shift (@_);    # config file name

	my @liste_toscan;

	debug("-------- begin of config file -------------\n");

	open( INIT, "<", $config )
	  or die "abort : can not open config file $config : $!\n";

	# for multi_line exclude_suffix : to detect first call
	my $first_exclude = 1;
	while ( my $line = <INIT> ) {
		chomp $line;

		$line =~ s/#.*//;    # skip comments
		remove_trailing_spaces( \$line );
		$line =~ s?\\?/?g;    # replace \ by / (for windows)

		next unless length($line);    # skip blank lines

		#debug( "config line = $line\n");

		if ( $line =~ m/\@\@/ ) {

			# skip macros lines
			###################
			debug("skipped macro in line $. : $line\n");
			next;
		}
		elsif ( $line =~ m/^=/ ) {

			# begin with a = : a directory to scan without recurse
			######################################################
			$line =~ s/^=\s*//;

			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);
			my $masqb   = pop (@tab);
			my $masq    = resolv_alias( $., $masqb );

			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING: skipped config file $line (line $.) : $elem already in config file\n";
				}
				elsif ( -d $elem ) {
					push ( @liste_toscan, $elem );
					$Toscan{$elem} = $masq;
					if ($flagdir) {
						$Onlydir{$elem} = 1;
						debug("toscan without sub-dir : $elem masq $masq\n");
					}
					else {
						$Onlythis{$elem} = 1;
						debug("toscan only this inode : $elem masq $masq\n");
					}
				}
				elsif ( -e _ ) {
					warn
"WARNING: skipped config file $elem (line $.) : $elem not a directory\n";
				}
				else {
					warn
"WARNING: skipped config file $line (line $.) : $elem does not exists\n";
				}
			}    # foreach my $elem
		}
		elsif ( $line =~ m/:=/ ) {

			# a configuration line
			######################
			my ( $key, $val ) = split ( /:=/, $line, 2 );
			remove_trailing_spaces( \$val );

			if ( $key =~ m/^database/ ) {
				debug("found database : $val\n");
				$database = $val;
			}
			elsif ( $key =~ m/^verbose/ ) {
				if ( !defined $Verbose ) {
					debug("found verbose : $val\n");
					$Verbose = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^warn_dead_symlinks/ ) {
				if ( !defined $Warn_dead_symlinks ) {
					debug("found warn_dead_symlinks : $val\n");
					$Warn_dead_symlinks = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^report_full_newdel/ ) {
				if ( !defined $Report_full_newdel ) {
					debug("found report_full_newdel : $val\n");
					$Report_full_newdel = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^warn_missing_file/ ) {
				if ( !defined $Warn_missing_file ) {
					debug("found warn_missing_file : $val\n");
					$Warn_missing_file = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}

			}
			elsif ( $key =~ m/^report_url/ ) {
				debug("found report_url : $val\n");
				if ( lc($val) eq "stdout" ) {
					$Report_url = "stdout";
				}
				elsif ( lc($val) eq "stderr" ) {
					$Report_url = "stderr";
				}
				elsif ( lc($val) eq "null" ) {
					$Report_url = 'null';
				}
				else {
					warn
"WARNING: skipped config file $line (line $.) : bad value $val\n";
				}
			}
			elsif ( $key =~ m/^exclude_suffix/ ) {
				if ( $first_exclude && defined(@ExSufx) ) {
					ignore_directive( $line, $. );
				}
				else {
					$first_exclude = 0;
					debug("found exclude_suffix : $val\n");

					# enable multi-line syntaxe
					push ( @ExSufx, split ( ' ', $val ) );
				}
			}
			elsif ( $key =~ m/^timing/ ) {
				if ( !defined $Timing ) {
					debug("found timing : $val\n");
					$Timing = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			elsif ( $key =~ m/^running_files/ ) {
				if ( !defined $Running ) {
					debug("found running_files : $val\n");
					$Running = decode( $val, $line, $. );
				}
				else {
					ignore_directive( $line, $. );
				}
			}
			else {
				warn
"WARNING: skipped config file $line (line $.) : unknown directive $key\n";
			}
		}
		elsif ( $line =~ m/=/ ) {

			# is an alias definition
			########################
			remove_all_spaces( \$line );
			my ( $key, $val ) = split ( /=/, $line );
			my $realval = resolv_alias( $., $val );
			$alias{$key} = $realval;
			debug("alias $key -> $realval\n");
		}

		elsif ( $line =~ m/^!/ ) {

			# begin with ! : it is an exception
			###################################
			$line =~ s/^!\s*//;

			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);          # and ignore
			my $masq    = pop (@tab);          # and ignore
			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING: skipped config file $line (line $.) : exception $elem already seen in config file\n";
				}
				else {
					debug("exception : $elem\n");
					$Toscan{$elem} = 0;
				}
			}
		}
		else {

			# directory or file to scan
			###########################
			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);
			my $masqb   = pop (@tab);
			my $masq    = resolv_alias( $., $masqb );

			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING: skipped config file $line (line $.) : $elem already in config file\n";
				}
				elsif ( -e $elem ) {
					push ( @liste_toscan, $elem );
					debug("toscan : $elem masq $masq\n");
					$Toscan{$elem} = $masq;
				}
				elsif ($Warn_missing_file) {
					warn
"WARNING: skipped config file $line (line $.) : $elem does not exists\n";
				}
			}    # foreach my $elem
		}
	}    # foreach my $line
	close(INIT);

	# set default values if not founds
	$Verbose            = 0 if ( !defined $Verbose );
	$Warn_dead_symlinks = 0 if ( !defined $Warn_dead_symlinks );
	$Report_full_newdel = 0 if ( !defined $Report_full_newdel );
	$Warn_missing_file  = 0 if ( !defined $Warn_missing_file );
	$Timing             = 0 if ( !defined $Timing );
	$Running            = 0 if ( !defined $Running );

# add some files for internal check : this program, the graphical interface, the config file
#my $masq = resolv_alias( $., "all" );
	my $masq = $alias{"all"};

	my @list = get_names();

	foreach my $elem (@list) {
		$Toscan{$elem} = $masq;
		push ( @liste_toscan, $elem );
	}

	$Toscan{$config} = $masq;
	push ( @liste_toscan, $config );

	# a summary of used config
	print_config();

	debug("-------- end of config file -------------\n");

	return @liste_toscan;
}
#######################################################
sub split_acl($) {
	my $acl = shift (@_);

	my @tab_acl = split ( /$sepacl/, $acl );
	my @text;
	foreach my $acl (@tab_acl) {
		my ( $sid, $mask ) = split ( '=', $acl );
		my @rights;
		Win32::FileSecurity::EnumerateRights( $mask, \@rights );
		push ( @text, "$sid=" . join ( ',', @rights ) );
	}
	return @text;
}
#######################################################
# display
# convert fields in a human way
sub display($@) {
	my $elem = shift (@_);
	my @info = @_;

	if ( $elem == $id{'device'} ) {

		# device is in hexa, ex 309 == major 3, minor 9 == /dev/hda9
		return sprintf "%lx", $info[$elem];
	}
	elsif ( $elem == $id{'file_mode'} ) {

		# perm is in octal
		return sprintf "%lo", $info[$elem];
	}
	elsif ( ( $elem == $id{'acl'} ) and ($acl_exist) ) {

		# uid on windows are acl
		return join ( ';', split_acl( $info[$elem] ) );
	}
	elsif (( $elem == $id{'atime'} )
		or ( $elem == $id{'mtime'} )
		or ( $elem == $id{'ctime'} ) )
	{

		# dates
		if ( $info[$elem] ) {
			return gmtime( $info[$elem] );
		}
		else {
			return 0;
		}
	}
	else {
		return $info[$elem];
	}
}

#######################################################
# display_all
# convert in a human way all the record
sub display_all($) {
	my $record = shift (@_);
	my @fileinfo = split ( /$sepmeta/, $record );

	my $output;

	my $max = scalar @fileinfo;
	for ( my $i = 0 ; $i < $max ; $i++ ) {
		$output .= display( $i, @fileinfo ) . "$sep";
	}
	return $output;
}
#######################################################
# md5sum :
# build a MD5 checksum on the given file
sub md5sum($) {
	my $name = shift (@_);    # file name

	# !!! the Digest module is not in standard
	# so we can not write Digest->new("MD5");
	my $ctx = Digest::MD5->new();

	if ( open( FILE, "<", "$name" ) ) {
		binmode(FILE);
		$ctx->addfile(*FILE);
		close(FILE);
		my $sum = $ctx->b64digest;
		return $sum;
	}
	else {
		warning("(md5sum) can not open $name for checksum: $!\n");
		return 0;
	}
}
#######################################################
# md5sum :
# build a MD5 checksum on the given file
sub sha1sum($) {
	my $name = shift (@_);    # file name

	# !!! the Digest module is not in standard
	# so we can not write Digest->new("MD5");
	my $ctx = Digest::SHA1->new();

	if ( open( FILE, "<", "$name" ) ) {
		binmode(FILE);
		$ctx->addfile(*FILE);
		close(FILE);
		my $sum = $ctx->b64digest;
		return $sum;
	}
	else {
		warning("(md5sum) can not open $name for checksum: $!\n");
		return 0;
	}
}
#######################################################
sub winacl($) {
	my $filename = shift (@_);

	my $acl = 0;
	my %perm;
	if ($acl_exist) {
		my @acl;
		if ( Win32::FileSecurity::Get( $filename, \%perm ) ) {
			while ( my ( $name, $mask ) = each %perm ) {
				$name =~ s/ /_/g;
				push ( @acl, "$name=$mask" );
			}
			$acl = join ( $sepacl, @acl );
		}
		else {
			warning( "Error #" . int($!) . ": $!" );
		}
	}
	return $acl;
}
#######################################################
# file_info :
# build a complete info on the given file
sub file_info($$) {
	my $name = shift (@_);
	my $masq = shift (@_);

	# for afick-tk progress text
	progress("$name") if ($Progress);

	debug( $name . $sep );

	my @fileinfo = lstat($name);

	my $sum;
	if ( -f $name ) {
		if ( ( $masq =~ m/h/ ) or ( $masq =~ m/5/ ) ) {
			$sum = md5sum($name);
		}
		elsif ( $masq =~ m/1/ ) {
			$sum = sha1sum($name);
		}
		else {
			$sum = 0;
		}
	}
	else {
		$sum = 0;
	}

	# Each file gets a pipe-delimited entry, with format as follows:
	#
	# field 0 = checksum hash
	# field 1 = device number of filesystem
	# field 2 = inode number
	# field 3 = file mode (type/permissions)
	# field 4 = number of hard links to file
	# field 5 = uid of the file
	# field 6 = gid of the file
	# field 7 = total size of file, in bytes
	# field 8 = actual number of blocks allocated
	# field 9 = last access time in seconds since the epoch
	# field 10 = last modify time (since epoch)
	# field 11 =  inode change time
	#
	# See the perl "stat" command or man stat() for more information.

	# default value is 0
	my (
		$device, $inode, $mode,  $link,  $uid, $gid,
		$size,   $bloc,  $atime, $mtime, $ctime
	  )
	  = ( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );

	$device = $fileinfo[0] if ( $masq =~ m/d/ );
	$inode  = $fileinfo[1] if ( $masq =~ m/i/ );

	# file mode are different on unix and windows
	# user/group/other and acl
	# but we use mode to get file type, so keep this field as it
	$mode = $fileinfo[2] if ( $masq =~ m/p/ );
	$link = $fileinfo[3] if ( $masq =~ m/n/ );
	$uid  = $fileinfo[4] if ( $masq =~ m/u/ );

	#$gid   = $fileinfo[5]  if ( $masq =~ m/g/ );
	if ( $masq =~ m/g/ ) {
		if ( $^O =~ m/^MSWin/ ) {

			# gid is unused on windows
			# it is replaced by the acess control list
			$gid = winacl($name);
		}
		else {

			# unix
			$gid = $fileinfo[5];
		}
	}

	$size  = $fileinfo[7]  if ( $masq =~ m/s/ );
	$atime = $fileinfo[8]  if ( $masq =~ m/a/ );
	$mtime = $fileinfo[9]  if ( $masq =~ m/m/ );
	$ctime = $fileinfo[10] if ( $masq =~ m/c/ );
	$bloc  = $fileinfo[12] if ( $masq =~ m/b/ );

	my $entry = join ( $sep,
		$sum, $device, $inode, $mode,  $link,  $uid,
		$gid, $size,   $bloc,  $atime, $mtime, $ctime );

	debug( display_all($entry) . "\n" );

	if ($Running) {

		# !! seems to be a problem with -M/-C on dangling links
		if ( $masq =~ m/m/ ) {
			my $age = -M $name;
			warning("$name as been modified during the program run\n")
			  if ( ( defined $age ) and ( $age < 0 ) );
		}
		if ( $masq =~ m/c/ ) {
			my $age = -C $name;
			warning("inode of $name as been modified during the program run\n")
			  if ( ( defined $age ) and ( $age < 0 ) );
		}
	}

	return $entry;
}
#######################################################
sub hash_acl(@) {
	my @tab = @_;
	my %acl;

	foreach my $elem (@tab) {
		my ( $sid, $acl ) = split ( /=/, $elem );
		$acl{$sid} = $acl;
	}
	return %acl;
}
#######################################################
sub compare_acl($$) {
	my $acl1 = shift (@_);
	my $acl2 = shift (@_);

	my %acl1 = hash_acl( split_acl($acl1) );
	my %acl2 = hash_acl( split_acl($acl2) );

	foreach my $key ( keys %acl1 ) {
		if ( exists $acl2{$key} ) {
			if ( $acl1{$key} ne $acl2{$key} ) {

				# changed acl
				report_detailed( "change acl $key", $acl1{$key}, $acl2{$key} );
			}
			delete $acl2{$key};
		}
		else {

			# deleted acl
			report_detailed( "deleted acl $key", $acl1{$key} );
		}
	}

	# we deleted all common keys on acl2
	# so we just have new acl now
	foreach my $key ( keys %acl2 ) {

		# new acl
		report_detailed( "new acl $key", $acl2{$key} );
	}

}
#######################################################
# compare_entry :
# compare 2 file infos and print the difference
sub compare_entry($$$) {
	my $name   = shift (@_);
	my $entry1 = shift (@_);    # old value
	my $entry2 = shift (@_);    # new value

	my @fileinfo1 = split ( /$sepmeta/, $entry1 );
	my @fileinfo2 = split ( /$sepmeta/, $entry2 );

	my $i;
	my $max = scalar @fileinfo1;
	for ( $i = 0 ; $i < $max ; $i++ ) {
		if ( $fileinfo1[$i] ne $fileinfo2[$i] ) {

			# !!!! for windows acl, it is long and difficult to read
			# so a special report is necessary
			if ( ( $i == $id{'acl'} ) and $acl_exist ) {
				compare_acl( $fileinfo1[$i], $fileinfo2[$i] );
			}
			else {

				# general case
				report_detailed(
					"$field[$i]",
					display( $i, @fileinfo1 ),
					display( $i, @fileinfo2 )
				);
			}
		}    # if fileinfo ne
	}    # for
}

#######################################################
# is_exception :
# return true if the file name belong exception list
sub is_exception($) {
	my $elem = shift (@_);
	return ( exists( $Toscan{$elem} ) and ( $Toscan{$elem} eq "0" ) );
}
#######################################################
# parcours :
# the general sub to scan a directory (recursive)
# should be called by config list entries
sub parcours($$$$) {
	my $rep     = shift (@_);    # file or directory to scan
	my $traite  = shift (@_);    # sub to call
	my $masq    = shift (@_);    # herited masq
	my $newflag = shift (@_);    # herited new flag

	#debug ("parcours : $rep $masq\n");

	my @liste;

	# first build a list of file
	if ( -d $rep ) {
		return if ( $rep =~ /\/\.\.?$/ );

		if ( exists $Toscan{$rep} ) {
			$masq = $Toscan{$rep};

			debug("change masq to $masq for $rep\n");
		}
		$traite->( $rep, $masq, $newflag );

		# equal : do not scan into the directory
		return if ( exists $Onlythis{$rep} );

		if ( !-l $rep ) {

			if ( !opendir( DIR, $rep ) ) {
				warning("(parcours) can not open directory $rep : $!\n");
				return;
			}
			my $elem;
			while ( defined( $elem = readdir(DIR) ) ) {
				next if $elem =~ /^\.\.?$/;

				#debug( "readdir : $elem");

				# equal : do not scan sub-dir
				next
				  if (  ( exists $Onlydir{$rep} )
					and ( -d $rep . "/" . $elem ) );

				# special case : / to avoid //
				if ( $rep eq "/" ) {
					push ( @liste, '/' . $elem );
				}
				else {
					push ( @liste, $rep . '/' . $elem );
				}
			}
			closedir(DIR);
		}
		else {

			# on ne suit pas les liens sur les repertoires
			debug("$rep skipped (link on directory)\n");
		}
	}
	else {
		push ( @liste, $rep );
	}

	parcours2( $rep, $traite, $masq, $Newfiles{$rep}, @liste );
}
#######################################################

sub parcours2($$$$@) {
	my $rep     = shift (@_);    # file or directory to scan
	my $traite  = shift (@_);    # sub to call
	my $masq    = shift (@_);    # herited masq
	my $newflag = shift (@_);    # herited new flag
	my @liste   = @_;

	foreach my $fic (@liste) {
		next if $fic =~ /^\.\.?$/;
		next if ( is_exception($fic) );
		next if ( test_exclude_suffix($fic) );

		# heritage, set to all by default (ex : -l action)
		my $lmasq = ( $masq or $alias{"all"} );

		# surcharge
		if ( exists $Toscan{$fic} ) {
			$lmasq = $Toscan{$fic};

			#debug ("change masq to $lmasq for $fic");
		}

		debug("parcours2 : $fic\n");

		if ( -d $fic ) {
			parcours( $fic, $traite, $lmasq, $newflag );
		}    # end directory
		elsif ( -l $fic ) {
			my $whatlink = readlink($fic);
			my $abs_whatlink;

			# is the link an absolute path ?
			if ( $whatlink !~ m/^\// ) {

				# absolute path for check
				$abs_whatlink = $rep . "/" . $whatlink;
			}
			else {
				$abs_whatlink = $whatlink;
			}
			if ( !-e $abs_whatlink ) {

				# real path for display
				$Dangling{$fic} = $whatlink;
			}
			$traite->( $fic, $lmasq, $Newfiles{$rep} );
		}    # end link
		else {
			$traite->( $fic, $lmasq, $Newfiles{$rep} );
		}    # end normal file
	}    # foreach
}
#######################################################
# wanted_create
# process a file in create mode
sub wanted_create($$) {
	my $name = shift (@_);
	my $masq = shift (@_);

	$Finded{$name} = 1;

	my $entry = file_info( $name, $masq );

	$Hashfile{$name} = $entry;
}
#######################################################
# end_report
sub end_report() {

	# Generate a MD5 for the dbm file.
	my $sum = md5sum( $database . $database_ext );

	info( $comment x 65 . "\n" );
	info("MD5 hash of $database => $sum\n\n");
	write_control($sum);
}
#######################################################
# create (base liste)
# create a new database (empty the existing one if exists)
sub create($@) {
	my $database = shift (@_);
	my @listerep = @_;
	tie( %Hashfile, $Dbm, $database, O_WRONLY | O_CREAT, 0666 )
	  or die "abort (create) : can not open $Dbm $database : $!";

	# delete all previous keys
	%Hashfile = ();

	foreach my $elem (@listerep) {

		# to skip already scanned files
		next if ( exists $Finded{$elem} );

		parcours( $elem, \&wanted_create, "", undef );
	}
	my $nbscan = scalar( keys %Finded );
	print_dangling(1);
	untie %Hashfile;

	report("\n");
	info("Hash database created successfully. $nbscan files entered.\n");

	end_report();
}
#######################################################
# wanted_update
# process a file in update mode
sub wanted_update($$$) {
	my $name    = shift (@_);
	my $masq    = shift (@_);
	my $newflag = shift (@_);

	$Finded{$name} = 1;

	my $entry = file_info( $name, $masq );

	if ( $Hashfile{$name} ) {

		# test for changes
		my $old_entry = $Hashfile{$name};
		if ( $old_entry ne $entry ) {

			# keep values to compare
			$Oldval{$name} = $old_entry;
			$Newval{$name} = $entry;
			$Nbmod++;
			$Hashfile{$name} = $entry if ($Update);
		}
	}
	else {

		# new entry
		if ( defined $newflag ) {
			$Newfiles{$name} = $newflag + 1;
		}
		else {
			$Newfiles{$name} = 1;
		}

		$Newval{$name} = $entry;

		#debug ("$name heritage : $newflag value $Newfiles{$name}\n");
		$Hashfile{$name} = $entry if ($Update);
	}
}    # end of wanted
#######################################################
# update :
# compare or update a database with the system
sub update($@) {
	my $database = shift (@_);
	my @listerep = @_;
	tie( %Hashfile, $Dbm, $database, O_RDWR, 0666 )
	  or die "abort (update) : can not open $Dbm $database : $!";

	my ( $old_version, $run_date ) = calc_control();

	# for afick-tk progress bar
	# guess the number of file does not change too much
	my $total = scalar( keys %Hashfile );
	progress("total $total") if ($Progress);

	# scan file list
	foreach my $elem (@listerep) {

		# skip already scanned
		next if ( exists $Finded{$elem} );

		parcours( $elem, \&wanted_update, "", undef );
	}

	# analysis
	my $nbchange = $Nbmod;
	my $nbdelete = 0;
	my $nbnew    = 0;
	if ( !$File ) {

		# display a general message
		if ( $nbchange and defined($run_date) ) {
			info( "changes since last run on "
				  . gmtime($run_date)
				  . " with afick version "
				  . $old_version
				  . "\n" );
		}

		# look for new files
		$nbnew = print_new(0);
		$Nbmod += $nbnew;

		# look for deleted files
		$nbdelete = print_delete(0);
		$Nbmod += $nbdelete;
	}
	print_changed(0);
	print_dangling(0);

	report("\n");
	info("detailed changes\n") if ($Nbmod);
	print_new(1);
	print_delete(1);
	print_changed(1);
	print_dangling(1);

	untie %Hashfile;

	report("\n");
	my $nbscan   = scalar( keys %Finded );
	my $dangling = scalar( keys %Dangling );
	my $text     =
"$nbscan files scanned, $Nbmod changed (new : $nbnew; delete : $nbdelete; changed : $nbchange; dangling : $dangling)\n";
	if ($Update) {
		info("Hash database updated successfully : $text");
	}
	else {
		info("Hash database : $text");
	}

	end_report();

	# return a status
	return oct(
		join ( '',
			'0b',
			map { $_ ? 1 : 0; } ( $nbnew, $nbdelete, $nbchange, $dangling ) )
	);
}

#######################################################
# print_dangling
# print all danglink symbolic links
sub print_dangling($) {
	my $detailed = shift (@_);
	return unless ($Warn_dead_symlinks);

	debug("print_dangling begin detailed=$detailed\n");
	foreach my $key ( keys %Dangling ) {
		report_summary( "Dangling", $key );
		report_detailed( "linked_to", $Dangling{$key} ) if ($detailed);
	}
	debug("print_dangling end\n");
}

#######################################################
# print_dbm
# print the database content
sub print_dbm($) {
	my $database = shift (@_);
	tie( %Hashfile, $Dbm, $database, O_RDONLY, 0666 )
	  or die "abort (print_dbm) : Couldn't tie $Dbm file $database : $!";

	my ( $old_version, $run_date ) = calc_control();

	if ( defined $run_date ) {
		info( "last run on "
			  . gmtime($run_date)
			  . " with afick version "
			  . $old_version
			  . "\n" );
	}

	# display print format
	info( "format : file_type name" . $sep . join ( $sep, @field ) . "\n" );

	# to change to have same format as other ?
	foreach my $key ( sort keys %Hashfile ) {
		report(
			file_type($key) . $sep . display_all( $Hashfile{$key} ) . "\n" );
	}
	info( "number of file : " . scalar( keys %Hashfile ) . "\n" );
	untie %Hashfile;
}
#######################################################
# print a summary or detailed info on changes
sub print_changed($) {
	my $detailed = shift (@_);
	foreach my $key ( sort keys %Oldval ) {
		report_summary( "changed", $key );
		compare_entry( $key, $Oldval{$key}, $Newval{$key} ) if ($detailed);
	}
}
#######################################################
# get a forced ctime to display inode date for new files
sub ctime($) {
	my $key = shift (@_);
	my $ctime = get_ctime($key) || ( stat($key) )[10];
	return gmtime($ctime);
}
#######################################################
# we display as detailed info the inode last change date
sub print_new($) {
	my $detailed = shift (@_);

	# look for new files
	my $nb    = 0;
	my $nbnew = 0;
	foreach my $key ( sort keys %Newfiles ) {

		# report only if asked all ($Report_full_newdel) or first level
		if ($Report_full_newdel) {
			report_summary( "new", $key );
			report_detailed( "inode_date", ctime($key) ) if ($detailed);
		}
		elsif ( $Newfiles{$key} == 1 ) {
			if ($nb) {
				report_detailed( "number of new files", $nb );
			}
			report_summary( "new", $key );
			report_detailed( "inode_date", ctime($key) ) if ($detailed);
			$nb = 0;
		}
		else {
			$nb++;
		}

		$nbnew++;
	}
	if ($nb) {
		report_detailed( "number of new files", $nb );
	}
	return $nbnew;
}
#######################################################
sub parent_date($) {
	my $parent = shift (@_);
	my $mtime = get_mtime($parent) || ( stat($parent) )[9];
	return gmtime($mtime);
}
#######################################################
# we dispplay as info the date change of parent directory
sub print_delete($) {
	my $detailed = shift (@_);

	my %deleted;
	my $nb       = 0;
	my $nbdelete = 0;
	foreach my $key ( sort keys %Hashfile ) {
		if ( !exists( $Finded{$key} ) ) {

			# get directory for a file or parent directory for a directory
			my $path = dirname($key);

			if ($Report_full_newdel) {
				report_summary( "deleted", $key );
				report_detailed( "parent_date", parent_date($path) )
				  if ($detailed);
				$nb = 0;
			}
			else {

				#print "--- deleted $key $path\n";
				if ( exists $deleted{$path} ) {
					$deleted{$key} = $deleted{$path} + 1;
					$nb++;
				}
				else {
					$deleted{$key} = 1;
					if ($nb) {
						report_detailed( "number of deleted files", $nb );
					}
					$nb = 0;
					report_summary( "deleted", $key );
					report_detailed( "parent_date", parent_date($path) )
					  if ($detailed);
				}
			}
			delete $Hashfile{$key} if ( $detailed and $Update );
			$nbdelete++;
		}
	}
	if ($nb) {
		report_detailed( "number of deleted files", $nb );
	}
	return $nbdelete;
}
#######################################################
# print internal flag state
sub print_config() {
	debug("report_url=$Report_url\n");
	debug("Verbose=$Verbose\n");
	debug("Warn_dead_symlinks=$Warn_dead_symlinks\n");
	debug("Report_full_newdel=$Report_full_newdel\n");
	debug("Warn_missing_file=$Warn_missing_file\n");
	debug("exclude_suffix=@ExSufx\n");

	#debug("Output_format=$Output_format\n");
}
#######################################################
# usage
# print some help
sub usage($) {
	my $version = shift (@_);
	print <<EOHELP;

Usage: $0 [mandatory action] [other options]

Mandatory action (one and only one must be used) : 
 -i|--init                    initialize the hash.dbm database
 -k|--compare                 compare the hash.dbm database
 -l|--list fic1 .. fic2       check the files given in arg 
 -u|--update                  compare and update the hash.dbm database
 -p|--print                   print content of dbm database

Other options
 -c|--config_file file        configname of config file to use
 -f|--full_newdel             report full information for new or deleted directories
                               default: no
			       reverse : --nofull_newdel 
 -m|--missing_files           warn about files declared in config files 
                               which do not exists, 
			       default: no; 
			       reverse : --nomissing_files
 -r|--running_files           warn about "running" files : modified since program begin
                               default: no 
                               reverse: --norunning_files
 -s|--dead_symlinks           warn about dead symlinks 
                               default: no 
                               reverse: --nodead_symlinks
 -t|--timing		      Print timing statistics
				default: no
				reverse : --notiming
 -v|--verbose                 toggle verbose mode; 
                               default: no 
 			       reverse : --noverbose
 -P|--progress		      display the name of scanned files, to be used only by afick-tk
 -h|--help                    show this help page
 -V|--version                 show afick version
 -x|--exclude_suffix ext1 ext2        list of suffixes to ignore

Disclaimer:
This script is intended to provide a means for
detecting changes made to files, via a regular
comparison of MD5 hashes to an established baseline. 

Copyright (c) 2002 Eric Gerbier <gerbier\@users.sourceforge.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
EOHELP

}

#######################################################
sub version($) {
	my $version = shift (@_);
	print "\n";
	print "afick : another file integrity checker\nversion $version\n";
}
#######################################################
# read database previous checksum and compare with new one
sub read_control($) {
	my $newchecksum  = shift (@_);
	my $control_file = $database . $control_ext;

	my ( $oldchecksum, $old_version, $run_date );
	if ( open( CONTROL, "<", $control_file ) ) {
		$oldchecksum = <CONTROL>;
		chomp($oldchecksum);
		$old_version = <CONTROL>;
		chomp($old_version) if ( defined $old_version );
		$run_date = <CONTROL>;
		chomp($run_date) if ( defined $run_date );
		close(CONTROL);
		if ( $oldchecksum ne $newchecksum ) {
			warn
"WARNING: internal change in afick database $database (see below)\n";
		}
	}
	else {
		warning(
			"(read_control) can not read control file $control_file : $!\n");
	}
	return ( $old_version, $run_date );
}
#######################################################
# write database checksum for next run
sub write_control ($) {
	my $checksum     = shift (@_);                 # the string to write
	my $control_file = $database . $control_ext;

	# to avoid Tainted warnings
	$control_file = $1 if ( $control_file =~ /(.+)/ );

	if ( open( CONTROL, ">", $control_file ) ) {
		print CONTROL "$checksum\n";               # database checksum
		print CONTROL "$version\n";                # afick version
		print CONTROL time() . "\n";               # date of run
		close(CONTROL);
	}
	else {
		warning(
			"(write_control) can not write in control file $control_file : $!\n"
		);
	}
}
#######################################################
# compare old and new checksum for afick main components
sub control ($) {
	my $name = shift (@_);                         # file name to control

	my $newchecksum = md5sum($name);
	my $oldchecksum = get_md5($name);

	if ( !defined $oldchecksum ) {

		# first run
		return;
	}
	elsif ( $oldchecksum ne $newchecksum ) {
		warning("(control) afick internal change : $name (see below)\n");
		debug("oldchecksum=$oldchecksum newchecksum=$newchecksum\n");
	}
}
#######################################################
# check if afick change since last run
sub calc_control() {

	my @list = get_names();

	foreach my $elem (@list) {
		control($elem);    # afick.pl
	}

	control($configfile);    # afick-tk.pl

	my $database_check = md5sum( $database . $database_ext );
	return read_control($database_check);
}
########################## main #############################
# be careful : does not exist in standard on all linux
eval { require Digest::SHA1 };
if ($@) {
	$sha1_exist = 0;
}
else {
	$sha1_exist = 1;
}

if ( $^O =~ m/linux/ ) {
	$default_config_file = "/etc/afick.conf";
}
elsif ( $^O =~ m/^MSWin/ ) {
	$default_config_file = "xp.conf";

	# for acl list
	eval { require Win32::FileSecurity };
	if ($@) {
		warning("perl module Win32::FileSecurity not found : $@\n");

		# will work but without acl
		$acl_exist = 0;
	}
	else {
		$acl_exist = 1;
	}
}
else {
	$default_config_file = "afick.conf";
}

$| = 1;

Getopt::Long::Configure('no_ignore_case');
unless (
	GetOptions(
		"config_file|c=s"    => \$configfile,
		"full_newdel|f!"     => \$Report_full_newdel,
		"help|h"             => \$help,
		"init|i"             => \$init,
		"compare|k"          => \$compare,
		"list|l=s"           => \$list,
		"missing_files|m!"   => \$Warn_missing_file,
		"print|p"            => \$print,
		"progress|P"         => \$Progress,
		"running_files|r!"   => \$Running,
		"dead_symlinks|s!"   => \$Warn_dead_symlinks,
		"verbose|v!"         => \$Verbose,
		"version|V"          => \$Version,
		"exclude_suffix|x=s" => \$Sufx_list,
		"timing|t!"          => \$Timing,
		"update|u"           => \$Update
	)
  )
{
	usage($version);
	die "abort : incorrect option\n";
}

# -h : help
if ($help) {
	usage($version);
	exit;
}

# -V : version
if ($Version) {
	version($version);
	exit;
}

# compute exclude suffix list from arg line
if ($Sufx_list) {
	@ExSufx = split ( ? ?, $Sufx_list );
}

if ($configfile) {
	my ( $name, $path, undef ) = fileparse($configfile);
	$path = abs_path($path) . '/';    # convert in absolute path

	$configfile = $path . $name;      # convert in absolute path
	debug("config file : $configfile\n");
}
elsif ( -e $default_config_file ) {
	$configfile = $default_config_file;
}
else {
	usage($version);
	die
"abort : missing configfile name (-c flag) and default config file $default_config_file\n";
}
@listerep = read_config($configfile);

# no we should have a database name
if ( !$database ) {
	usage($version);
	die
"abort : missing database name in flag list (-b) and in config file $configfile\n";
}

my $return_value = 0;

# actions
if ($list) {

	# -l : liste
	#@listerep = trait_config( split ( ? ?, $liste ) );
	@listerep     = split ( ? ?, $list );
	$File         = 1;
	$return_value = update( $database, @listerep );
}
elsif ($print) {

	# -p : print
	print_dbm($database);
}
elsif ($init) {

	# -i : init
	create( $database, @listerep );
}
elsif ($Update) {

	# -u : update
	$return_value = update( $database, @listerep );
}
elsif ($compare) {

	# -k : check
	$return_value = update( $database, @listerep );
}
else {
	usage($version);
	die "abort : no action to do (-i, -u, -k, -l, -p) \n";
}

# timing info
if ($Timing) {
	my ( $user, $system, $cuser, $csystem ) = times();
	info("user time : $user; system time : $system\n");
}

exit $return_value;

__END__

=head1 NAME

afick : Another File Integrity Checker

=head1 SYNOPSIS

=head1 OPTIONS

=head1 DESCRIPTION

=cut
